@@ -7,7 +7,7 @@ Panda::XS - useful features and typemaps for XS modules.
 
 =cut
 
-our $VERSION = '2.0.0';
+our $VERSION = '2.1.0';
 require Panda::XSLoader;
 Panda::XSLoader::load();
 
@@ -166,7 +166,7 @@ Scalar/Array/Hash/IO reference based objects.
 
 =item T_OPTR
 
-=item YOUR_TYPE : T_OPTR([basetype=class], [refcnt], [backref], [wrapper=wrapper_class], [static_cast])
+=item YOUR_TYPE : T_OPTR([basetype=class], [refcnt], [backref], [wrapper=wrapper_class], [static_cast], [prevent_default_destroy])
 
 IV(SCALAR)-based object with C pointer attached.
 
@@ -195,51 +195,9 @@ In this case, set 'basetype' to the name of the most parent C++ class. See C<HOW
 
 See C<T_OEXT> for description of params.
 
-            =item T_OPTR_SHARED
-            
-            =item YOUR_TYPE : T_OPTR_SHARED([basetype=classname], [nocast=1])
-            
-            A typemap class for case when you want to store an object in a shared pointer. Shared pointer can be C++11's std::shared_ptr or
-            panda::shared_ptr (STL compatible shared pointer which works even without C++11).
-            
-            Note that if you want to use this typemap, your typemap type must be a shared pointer, not an object's type. Also you must set
-            a shared pointer to OUTPUT variable (RETVAL), and you will receive a shared pointer in INPUT args. For example:
-            
-                typedef shared_ptr<MyClass> MyClassSP;
-                
-                #typemap
-                MyClassSP T_OPTR_SHARED
-                
-                #XS
-                MyClassSP new (SV* CLASS, int val) {
-                    RETVAL = MyClassSP(new MyClass(val));
-                }
-                
-                int get_val (MyClassSP THIS) {
-                    RETVAL = THIS->val;
-                }
-                
-                void DESTROY (MyClassSP THIS)
-                
-            Also note that shared pointers are passed by value, not by pointer. See how we receive an INPUT arg 'THIS':
-            
-                int get_val (MyClassSP THIS)  // CORRECT
-                int get_val (MyClassSP* THIS) // WRONG!!!
-                
-            Unfortunately, when using shared pointers, class-methods-like XS syntax is unavailable:
-            
-                int MyClassSP::get_val () // WRONG!!!
-                
-            Because MakeMaker's ParseXS automatically rewrite it as
-            
-                int get_val (MyClassSP* THIS) // WRONG!!!
-            
-            Remember that you HAVE to define DESTROY function even if you don't have anything to do there.
-            Otherwise, shared pointer won't be released!
-
 =item T_OEXT
 
-=item YOUR_TYPE : T_OEXT([basetype=class], [refcnt], [backref], [wrapper=wrapper_class], [static_cast], [on_svdup=funcname])
+=item YOUR_TYPE : T_OEXT([basetype=class], [refcnt], [backref], [wrapper=wrapper_class], [static_cast], [on_svdup=funcname], [prevent_default_destroy])
 
 Extendable object with C pointer attached.
 
@@ -359,6 +317,11 @@ If you use wrapper
 
 See L<THREADED PERLS SUPPORT> for details.
 
+=item prevent_default_destroy
+
+Prevents T_OEXT/T_OPTR from inserting object deletion code to the end of XS DESTROY function. Use with caution!
+See L<DESTROY behaviour for T_OEXT and T_OPTR> for details.
+
 =back
 
 Special C variables for OUTPUT
@@ -485,6 +448,35 @@ and exactly like
 
 =back
 
+=head3 DESTROY behaviour for T_OEXT and T_OPTR
+
+These typemap classes always adds object deletion code to DESTROY function, even if your XS DESTROY function is non empty.
+
+The code is:
+
+=over
+
+=item If you use wrapper
+
+    delete THIS_wrapper;
+    
+=item If you use refcounting
+
+    xs::refcnt_dec(THIS); // which is THIS->release() for panda::RefCounted
+    
+=item If you use shared pointers
+
+    // Code to release shared pointer
+    
+=item Nothing of above
+
+    delete THIS;
+
+=back
+
+If you don't want T_OEXT / T_OPTR to insert that for you, set C<prevent_default_destroy> param for typemap class.
+
+Keep in mind that in this case you must by any means avoid memory leaks, i.e. do the same or similar.
 
 =head1 HOW TO
 
@@ -765,7 +757,6 @@ XS:
     
     void Milk::DESTROY () {
         xs::call_super(cv, &ST(0), items, G_DISCARD);
-        delete THIS;
     }
     
 =head2 CREATE AN XS CLASS WHICH IS SUITABLE FOR MULTIPLE IHERITANCE / C3 CLASS MIXIN
@@ -805,7 +796,6 @@ XS:
     
     void MyPlugin::DESTROY () {
         xs::call_next_maybe(cv, &ST(0), items, G_DISCARD);
-        delete THIS;
     }
     
 Perl:
@@ -1273,7 +1263,7 @@ bound to their SV object and should live in sync with it.
 The second solution of the problem above is shared pointers. It allows you to solve this problem even if you can't change sources
 of C++ classes.
 
-You can use either STL's std::shared_ptr or compiler-independent panda::shared_ptr which has the same API.
+You can use either STL's std::shared_ptr or panda::shared_ptr (STL compatible shared pointer which doesn't require C++11).
 
 T_OEXT/T_OPTR automatically detects if you use shared pointers. Everything will work as you expect.
 
@@ -1317,8 +1307,33 @@ T_OEXT/T_OPTR automatically detects if you use shared pointers. Everything will
         ...
         if (ix == 0) RETVAL = THIS->first_driver();
         else         RETVAL = THIS->second_driver();
-    }    
+    }
 
+You might have noted several things:
+
+=over
+
+=item DriverSP T_OEXT
+
+Typemap class is bound to shared pointer's type, not to an object's type.
+Therefore you must set a shared pointer to OUTPUT variable (RETVAL), and you will receive a shared pointer in INPUT args
+
+=item void DESTROY (DriverSP THIS)
+
+Shared pointers are passed by value, not by pointer. See how we receive an INPUT arg 'THIS':
+
+    const char* name (DriverSP THIS)  // CORRECT
+    const char* name (DriverSP* THIS) // WRONG!!!
+    const char* DriverSP::name ()     // WRONG!!!
+                
+The third example won't work because for ExtUtils::ParseXS it's just an alias for
+
+    const char* name (DriverSP* THIS)
+    
+:((
+
+=back
+            
 Using shared pointers introduces slight overheat, as it have to alloc additional memory per object. However, see next paragraph
 to find out how you can remove this overheat while still using shared pointers.
 
