@@ -1,6 +1,7 @@
 #pragma once
 #include <stdint.h>
 #include <stddef.h>
+#include <panda/cast.h>
 
 #ifdef CPP11X
 #  include <memory>
@@ -10,48 +11,52 @@ namespace panda {
 
 static long int void_refcnt = 1;
 
-template <typename T>
-struct HasRetain {
-    typedef char yes;
-    typedef char no[2];
+namespace {
+    template <typename T>
+    struct HasRetain {
+        typedef char yes;
+        typedef char no[2];
 
-    struct fallback { int retain; };
-    struct mixed_type: T, fallback {};
-    template < typename U, U > struct type_check {};
+        struct fallback { int retain; };
+        struct mixed_type: T, fallback {};
+        template < typename U, U > struct type_check {};
 
-    template < typename U > static no&  test( type_check< int (fallback::*), &U::retain >* = 0 );
-    template < typename U > static yes& test( ... );
+        template < typename U > static no&  test( type_check< int (fallback::*), &U::retain >* = 0 );
+        template < typename U > static yes& test( ... );
 
-    static const bool value = sizeof( yes ) == sizeof( test< mixed_type >( NULL ) );
-};
+        static const bool value = sizeof( yes ) == sizeof( test< mixed_type >( NULL ) );
+    };
 
-template <typename T>
-struct HasRelease {
-    typedef char yes;
-    typedef char no[2];
+    template <typename T>
+    struct HasRelease {
+        typedef char yes;
+        typedef char no[2];
 
-    struct fallback { int release; };
-    struct mixed_type: T, fallback {};
-    template < typename U, U > struct type_check {};
+        struct fallback { int release; };
+        struct mixed_type: T, fallback {};
+        template < typename U, U > struct type_check {};
 
-    template < typename U > static no&  test( type_check< int (fallback::*), &U::release >* = 0 );
-    template < typename U > static yes& test( ... );
+        template < typename U > static no&  test( type_check< int (fallback::*), &U::release >* = 0 );
+        template < typename U > static yes& test( ... );
 
-    static const bool value = sizeof( yes ) == sizeof( test< mixed_type >( NULL ) );
-};
+        static const bool value = sizeof( yes ) == sizeof( test< mixed_type >( NULL ) );
+    };
 
-template <typename T>
-struct IsRefCounted {
-    static const bool value = HasRelease<T>::value && HasRetain<T>::value;
-};
+    template <typename T>
+    struct IsRefCounted {
+        static const bool value = HasRelease<T>::value && HasRetain<T>::value;
+    };
+}
 
 class RefCounted {
 public:
-    void    retain  () const { ++_refcnt; }
-    void    release () const { if (--_refcnt <= 0) delete this; }
+    void    retain  () const { on_retain(); ++_refcnt; }
+    void    release () const { on_release(); if (--_refcnt <= 0) delete this; }
     int32_t refcnt  () const { return _refcnt; }
 protected:
     RefCounted () : _refcnt(0) {}
+    virtual void on_retain  () const {}
+    virtual void on_release () const {}
     virtual ~RefCounted () {}
 private:
     mutable int32_t _refcnt;
@@ -68,16 +73,16 @@ public:
     shared_ptr () : ptr(NULL) {}
 
     shared_ptr (T* pointer) : ptr(pointer) {
-        ptr->retain();
+        if (ptr) ptr->retain();
     }
 
     shared_ptr (const shared_ptr<T,true>& oth) : ptr(oth.ptr) {
-        ptr->retain();
+        if (ptr) ptr->retain();
     }
 
     template<class U>
     shared_ptr (const shared_ptr<U,true>& oth) : ptr(oth.get()) {
-        ptr->retain();
+        if (ptr) ptr->retain();
     }
 
     ~shared_ptr () {
@@ -129,8 +134,13 @@ public:
 
     explicit shared_ptr (T* pointer) {
         ptr = pointer;
-        refcnt = new long int;
-        *refcnt = 1;
+        if (ptr) {
+            refcnt = new long int;
+            *refcnt = 1;
+        } else {
+            refcnt = &void_refcnt;
+            refcnt_inc();
+        }
     }
 
     shared_ptr (const shared_ptr<T,false>& oth) : ptr(oth.ptr), refcnt(oth.refcnt) {
@@ -220,13 +230,13 @@ inline shared_ptr<T1,true> const_pointer_cast (const shared_ptr<T2,true>& shptr)
 
 template <typename T1, typename T2>
 inline shared_ptr<T1,false> dynamic_pointer_cast (const shared_ptr<T2,false>& shptr) {
-    if (T1* p = dynamic_cast<T1*>(shptr.get())) return shared_ptr<T1,false>(shptr, p);
+    if (T1* p = dyn_cast<T1*>(shptr.get())) return shared_ptr<T1,false>(shptr, p);
     return shared_ptr<T1,false>();
 }
 
 template <typename T1, typename T2>
 inline shared_ptr<T1,true> dynamic_pointer_cast (const shared_ptr<T2,true>& shptr) {
-    if (T1* p = dynamic_cast<T1*>(shptr.get())) return shared_ptr<T1,true>(p);
+    if (T1* p = dyn_cast<T1*>(shptr.get())) return shared_ptr<T1,true>(p);
     return shared_ptr<T1,true>();
 }
 
