@@ -16,8 +16,13 @@ extern "C" {
 #  define CPP11X
 #endif
 
+#include <panda/def.h>
+#include <panda/cast.h>
+#include <panda/traits.h>
 #include <panda/refcnt.h>
 
+#include <pthread.h>
+
 typedef SV OSV;
 typedef HV OHV;
 typedef AV OAV;
@@ -53,14 +58,6 @@ typedef IO OIO;
 
 // Threaded-perl helpers
 
-#ifndef THREAD_LOCAL
-#  ifdef _MSC_VER
-#    define THREAD_LOCAL __declspec( thread )
-#  else
-#    define THREAD_LOCAL __thread
-#  endif
-#endif
-
 #ifdef PERL_IMPLICIT_CONTEXT // define class member helpers for storing perl interpreter
 #  define mTHX      pTHX;
 #  define mTHXa(a)  aTHX(a),
@@ -78,14 +75,9 @@ enum next_t {
 };
 
 struct my_perl_auto_t { // per-thread interpreter to help dealing with pTHX/aTHX, especially for static initialization
-    operator PerlInterpreter*() const {
 #ifdef PERL_IMPLICIT_CONTEXT
-        static THREAD_LOCAL PerlInterpreter* _i = PERL_GET_THX;
-        return _i;
-#else
-        return NULL;
+    operator PerlInterpreter*() const { return PERL_GET_THX; }
 #endif
-    }
 };
 extern my_perl_auto_t my_perl;
 
@@ -144,31 +136,31 @@ inline int sv_payload_detach (pTHX_ SV* sv, payload_marker_t* marker) {
     return sv_unmagicext(sv, PERL_MAGIC_ext, marker ? marker : &sv_payload_default_marker);
 }
 
-inline void rv_payload_attach (pTHX_ const SV* rv, void* ptr, const payload_marker_t* marker = NULL) {
+inline void rv_payload_attach (pTHX_ SV* rv, void* ptr, const payload_marker_t* marker = NULL) {
     sv_payload_attach(aTHX_ SvRV(rv), ptr, marker);
 }
 
-inline void rv_payload_attach (pTHX_ const SV* rv, void* ptr, SV* obj, const payload_marker_t* marker = NULL) {
+inline void rv_payload_attach (pTHX_ SV* rv, void* ptr, SV* obj, const payload_marker_t* marker = NULL) {
     sv_payload_attach(aTHX_ SvRV(rv), ptr, obj, marker);
 }
 
-inline void rv_payload_attach (pTHX_ const SV* rv, SV* obj, const payload_marker_t* marker = NULL) {
+inline void rv_payload_attach (pTHX_ SV* rv, SV* obj, const payload_marker_t* marker = NULL) {
     sv_payload_attach(aTHX_ SvRV(rv), obj, marker);
 }
 
-inline bool rv_payload_exists (pTHX_ const SV* rv, const payload_marker_t* marker = NULL) {
+inline bool rv_payload_exists (pTHX_ SV* rv, const payload_marker_t* marker = NULL) {
     return sv_payload_exists(aTHX_ SvRV(rv), marker);
 }
 
-inline void* rv_payload (pTHX_ const SV* rv, const payload_marker_t* marker = NULL) {
+inline void* rv_payload (pTHX_ SV* rv, const payload_marker_t* marker = NULL) {
     return sv_payload(aTHX_ SvRV(rv), marker);
 }
 
-inline SV* rv_payload_sv (pTHX_ const SV* rv, const payload_marker_t* marker = NULL) {
+inline SV* rv_payload_sv (pTHX_ SV* rv, const payload_marker_t* marker = NULL) {
     return sv_payload_sv(aTHX_ SvRV(rv), marker);
 }
 
-inline int rv_payload_detach (pTHX_ const SV* rv, payload_marker_t* marker = NULL) {
+inline int rv_payload_detach (pTHX_ SV* rv, payload_marker_t* marker = NULL) {
     return sv_payload_detach(aTHX_ SvRV(rv), marker);
 }
 
@@ -177,6 +169,33 @@ inline SV* call_super       (pTHX_ CV* cv, SV** args, I32 items, I32 flags = 0)
 inline SV* call_next_method (pTHX_ CV* cv, SV** args, I32 items, I32 flags = 0) { return call_next(aTHX_ cv, args, items, NEXT_METHOD, flags); }
 inline SV* call_next_maybe  (pTHX_ CV* cv, SV** args, I32 items, I32 flags = 0) { return call_next(aTHX_ cv, args, items, NEXT_MAYBE, flags); }
 
+
+class XSBackref : public virtual panda::RefCounted {
+    public:
+        SV* perl_object;
+
+        void on_perl_dup (pTHX_ int32_t refcnt); /* should be called when interpreter is cloned */
+
+    protected:
+        XSBackref () : perl_object(NULL) {}
+
+        virtual void on_retain () const {
+            SvREFCNT_inc_simple_void(perl_object);
+        }
+
+        // XS DTOR typemap sets perl_object = NULL just before DTOR code, to avoid infinite loop
+        virtual void on_release () const {
+            SvREFCNT_dec(perl_object);
+        }
+
+        virtual ~XSBackref () {};
 };
 
+// interface to refcounted object. if user uses its own refcnt base class, he should add overloading for these functions
+inline int32_t refcnt_get (const panda::RefCounted* var) { return var->refcnt(); }
+inline void    refcnt_inc (const panda::RefCounted* var) { var->retain(); }
+inline void    refcnt_dec (const panda::RefCounted* var) { var->release(); }
+
+}
+
 #include <xs/xs-private.h>
